/******************************************************************************
*
* Copyright (C) 2017 Allegro DVT2.  All rights reserved.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* Use of the Software is limited solely to applications:
* (a) running on a Xilinx device, or
* (b) that interact with a Xilinx device through a bus or interconnect.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
* XILINX OR ALLEGRO DVT2 BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* Except as contained in this notice, the name of  Xilinx shall not be used
* in advertising or otherwise to promote the sale, use or other dealings in
* this Software without prior written authorization from Xilinx.
*
*
* Except as contained in this notice, the name of Allegro DVT2 shall not be used
* in advertising or otherwise to promote the sale, use or other dealings in
* this Software without prior written authorization from Allegro DVT2.
*
******************************************************************************/

#include "Sections.h"
#include "NalWriters.h"

#include <lib_common/SEI.h>

#include <assert.h>

static int getBytesOffset(AL_TBitStreamLite* pStream)
{
  return AL_BitStreamLite_GetBitsCount(pStream) / 8;
}

static int writeNalInBuffer(IRbspWriter* writer, uint8_t* buffer, AL_NalUnit* nal)
{
  AL_TBitStreamLite bitstream;
  AL_BitStreamLite_Init(&bitstream, buffer);

  nal->Write(writer, &bitstream, nal->param);

  return AL_BitStreamLite_GetBitsCount(&bitstream);
}

static void GenerateNal(IRbspWriter* writer, AL_TBitStreamLite* bitstream, AL_NalUnit* nal, AL_TStreamMetaData* pMeta, uint32_t uFlags)
{
  uint8_t tmpBuffer[ENC_MAX_HEADER_SIZE];

  int sizeInBits = writeNalInBuffer(writer, tmpBuffer, nal);

  int start = getBytesOffset(bitstream);
  FlushNAL(bitstream, nal->nut, nal->header, tmpBuffer, sizeInBits);
  int end = getBytesOffset(bitstream);

  AL_StreamMetaData_AddSection(pMeta, start, end - start, uFlags);
}

static void GenerateConfigNalUnits(IRbspWriter* writer, AL_NalUnit* nals, int nalsCount, AL_TBuffer* pStream)
{
  AL_TBitStreamLite bitstream;
  AL_BitStreamLite_Init(&bitstream, AL_Buffer_GetData(pStream));
  AL_TStreamMetaData* pMetaData = (AL_TStreamMetaData*)AL_Buffer_GetMetaData(pStream, AL_META_TYPE_STREAM);

  for(int i = 0; i < nalsCount; i++)
    GenerateNal(writer, &bitstream, &nals[i], pMetaData, SECTION_CONFIG_FLAG);
}

static SeiPrefixAPSCtx createSeiPrefixAPSCtx(AL_TSps* sps, AL_THevcVps* vps)
{
  SeiPrefixAPSCtx ctx = { sps, vps };
  return ctx;
}

static SeiPrefixCtx createSeiPrefixCtx(AL_TSps* sps, int initialCpbRemovalDelay, int cpbRemovalDelay, AL_TEncPicStatus const* pPicStatus, uint32_t uFlags)
{
  SeiPrefixCtx ctx = { sps, initialCpbRemovalDelay, cpbRemovalDelay, uFlags, pPicStatus };
  return ctx;
}

static SeiSuffixCtx createSeiSuffixCtx()
{
  SeiSuffixCtx ctx;
  Rtos_Memcpy(&ctx.uuid, SEI_SUFFIX_USER_DATA_UNREGISTERED_UUID, 16);
  return ctx;
}

static int getOffsetAfterLastSection(AL_TStreamMetaData* pMeta)
{
  if(pMeta->uNumSection == 0)
    return 0;

  AL_TStreamSection lastSection = pMeta->pSections[pMeta->uNumSection - 1];
  return lastSection.uOffset + lastSection.uLength;
}

static uint32_t generateSeiFlags(AL_TEncPicStatus const* pPicStatus)
{
  uint32_t uFlags = SEI_PT;

  if(pPicStatus->eType == SLICE_I)
  {
    uFlags |= SEI_BP;

    if(!pPicStatus->bIsIDR)
      uFlags |= SEI_RP;
  }

  return uFlags;
}

void GenerateSections(IRbspWriter* writer, Nuts nuts, const NalsData* nalsData, AL_TBuffer* pStream, AL_TEncPicStatus const* pPicStatus)
{
  AL_TStreamMetaData* pMetaData = (AL_TStreamMetaData*)AL_Buffer_GetMetaData(pStream, AL_META_TYPE_STREAM);

  if(pPicStatus->bIsFirstSlice)
  {
    AL_NalUnit nals[6];
    int nalsCount = 0;

    if(nalsData->shouldWriteAud)
      nals[nalsCount++] = AL_CreateAud(nuts.audNut, pPicStatus->eType);

    if(pPicStatus->bIsIDR)
    {
      if(writer->WriteVPS)
        nals[nalsCount++] = AL_CreateVps(nalsData->vps);

      nals[nalsCount++] = AL_CreateSps(nuts.spsNut, nalsData->sps);
    }

    if(pPicStatus->eType == SLICE_I)
      nals[nalsCount++] = AL_CreatePps(nuts.ppsNut, nalsData->pps);

    SeiPrefixAPSCtx seiPrefixAPSCtx;
    SeiPrefixCtx seiPrefixCtx;

    if(nalsData->seiData)
    {
      assert(nalsData->seiFlags != SEI_NONE);
      uint32_t const uFlags = generateSeiFlags(pPicStatus) & nalsData->seiFlags;

      if(uFlags & SEI_BP && writer->WriteSEI_ActiveParameterSets)
      {
        seiPrefixAPSCtx = createSeiPrefixAPSCtx(nalsData->sps, nalsData->vps);
        nals[nalsCount++] = AL_CreateSeiPrefixAPS(&seiPrefixAPSCtx, nuts.seiPrefixNut);
      }

      seiPrefixCtx = createSeiPrefixCtx(nalsData->sps, nalsData->seiData->initialCpbRemovalDelay, nalsData->seiData->cpbRemovalDelay, pPicStatus, uFlags);
      nals[nalsCount++] = AL_CreateSeiPrefix(&seiPrefixCtx, nuts.seiPrefixNut);
    }

    for(int i = 0; i < nalsCount; i++)
      nals[i].header = nuts.GetNalHeader(nals[i].nut, nals[i].idc);

    GenerateConfigNalUnits(writer, nals, nalsCount, pStream);
  }

  AL_TStreamPart* pStreamParts = (AL_TStreamPart*)(AL_Buffer_GetData(pStream) + pPicStatus->uStreamPartOffset);

  for(int iPart = 0; iPart < pPicStatus->iNumParts; ++iPart)
    AL_StreamMetaData_AddSection(pMetaData, pStreamParts[iPart].uOffset, pStreamParts[iPart].uSize, 0);

  AL_TBitStreamLite bs;
  AL_BitStreamLite_Init(&bs, AL_Buffer_GetData(pStream));
  int offset = getOffsetAfterLastSection(pMetaData);
  AL_BitStreamLite_SkipBits(&bs, offset * 8);

  if(nalsData->shouldWriteFillerData && pPicStatus->iFiller)
  {
    WriteFillerData(&bs, nuts.fdNut, nuts.GetNalHeader(nuts.fdNut, 0), pPicStatus->iFiller);
    AL_StreamMetaData_AddSection(pMetaData, offset, pPicStatus->iFiller, 0);
  }

  if(pPicStatus->bIsLastSlice)
  {
    if(nalsData->seiData)
    {
      if(nalsData->seiFlags & SEI_EOF)
      {
        SeiSuffixCtx seiSufficCtx = createSeiSuffixCtx();
        AL_NalUnit nal = AL_CreateSeiSuffix(&seiSufficCtx, nuts.seiSuffixNut);
        nal.header = nuts.GetNalHeader(nal.nut, nal.idc);
        GenerateNal(writer, &bs, &nal, pMetaData, 0);
      }
    }
    AL_StreamMetaData_AddSection(pMetaData, 0, 0, SECTION_END_FRAME_FLAG);
  }

  if(pPicStatus->bIsIDR)
    AddFlagsToAllSections(pMetaData, SECTION_SYNC_FLAG);
}

